var gulp = require('gulp'),
    assert = require("assert"),
    runSequence = require('run-sequence'),
    fs = require('fs'),
    Path = require('path'),
    basset = require('../index.js');


var buildDir = "public/builds";

describe("Basset", function() {

  describe("build", function() {

    beforeEach(function() {
      var manifest = Path.join(buildDir, 'manifest.json')
      if (fs.existsSync(manifest)) {
        fs.unlinkSync(manifest);
      }

      basset.scan(require("./collections.json"));
    })

    it("should build manifest.json", function(done) {

      runSequence('build:main', function() {
        var collectionFile = Path.join(buildDir, 'manifest.json')

        assert(fs.existsSync(collectionFile));

        assert.equal(fs.readFileSync(collectionFile)+"", "{\n\
  \"main\": {\n\
    \"local\": {\n\
      \"js\": {\n\
        \"main/main.js\": \"main/main/main-c4238edcaf585158686a3ecf88f610d5d0246773.js\"\n\
      },\n\
      \"css\": {}\n\
    }\n\
  }\n\
}");
        done()
      });
    })

    it("builds files with hashed names", function() {
      var builtFile = Path.join(buildDir, "main/main/main-c4238edcaf585158686a3ecf88f610d5d0246773.js");

      assert(fs.existsSync(builtFile));
    })

    it('ads angular dependency injection', function(done) {

      basset.scan({
        "collections": {
          "ng": {
            "js": [
              "ng.js"
            ]
          }
        }
      });

      runSequence('production', 'build:ng', function() {
        var contents = fs.readFileSync(Path.join(buildDir, "ng-55b41af3002e11e864a74175e312a20217c6eaf4.js"));

        assert.equal('var app=angular.module("ang",[]);app.controller("Controller",["$scope",function(){console.log("works")}]),app.config(["$rootScope",function(){console.log("works")}]),app.run(["$rootScope",function(){console.log("works")}]);', contents.toString());

        done()
      });
    })

  });
});
